@@ -3,12 +3,13 @@ import json
 import os
 import difflib
 import re
+import codecs
 
 from web3 import Web3
 import requests
 
 from peth.eth.sigs import ERC20Signatures, Signature
-from peth.eth.utils import get_4byte_sig, sha3_256
+from peth.eth.utils import selector_to_sigs, sha3_256, SelectorDatabase, convert_value, hex2bytes, guess_calldata_types
 from peth.eth.bytecode import Code
 from peth.eth.opcodes import OpCode
 from peth.core.peth import Peth
@@ -17,6 +18,7 @@ from peth.util.graph import ContractRelationGraph
 
 from . import config
 from .config import chain_config, contracts_config, OUTPUT_PATH
+from .log import logger, logging
 
 
 class PethConsole(cmd.Cmd):
@@ -114,6 +116,11 @@ class PethConsole(cmd.Cmd):
         Toggle debug flag. Once on, the console will raise exceptions instead of catching them.
         """
         self._debug = not self._debug
+        if self._debug:
+            logger.setLevel(logging.DEBUG)
+        else:
+            logger.setLevel(logging.INFO)
+            
         print("debug set to", self._debug)
 
     def do_chain(self, arg):
@@ -184,17 +191,85 @@ class PethConsole(cmd.Cmd):
 
     def do_4byte(self, arg):
         """
-        4byte <hex_sig> : query text signature in 4byte database.
+        4byte <selector> : query text signature in https://sig.eth.samczsun.com/.
+        4byte <text> : query text signature which includes such text in local database.
         """
         if not arg:
-            print("4byte <hex_sig> :query text signature in 4byte database.")
+            print("4byte <hex_sig> :query text signature in https://sig.eth.samczsun.com/.")
             return
 
-        sigs = get_4byte_sig(arg)
-        if sigs:
-            print('\n'.join(sigs))
+        db = SelectorDatabase.get()
+
+        if re.match('^[0-9A-Fa-fXx]*$', arg): # selector
+            sigs = db.get_sig_from_selector(arg, False, True)
+            if sigs:
+                print('\n'.join(sigs))
+            else:
+                print("Selector not found in https://sig.eth.samczsun.com/.")
         else:
-            print("Not found in 4byte.directory.")
+            ret = db.get_sig_from_text(arg)
+            full_match = None
+            if ret:
+                for selector, sigs in ret:
+                    if arg in sigs:
+                        full_match = (selector, ', '.join(sigs))
+                    print("0x%s %s" % (selector, ', '.join(sigs)))
+
+            print("%d item(s) found in 4byte.json." % len(ret))
+            if full_match:
+                print("Full match: 0x%s %s" % full_match)
+
+    def do_abi_encode(self, arg):
+        """
+        abi_encode <sig> <args1> <args2> ...
+        eg: abi_encode func(uint256,address) 1 0xEEEEE...
+        """
+        args = arg.split()
+        sig = args[0]
+        s = Signature.from_sig(sig)
+
+        if len(args) == 1: # Only selector.
+            print('0x' + s.selector.hex())
+            return
+
+        args = args[1:]
+        args = list(map(convert_value, args))
+        print('0x' + s.encode_args(args, True).hex())
+
+    def do_abi_decode(self, arg):
+        """
+        abi_encode <hex>
+        abi_encode <hex> <sig>
+        """
+        args = arg.split()
+        hexdata = args[0]
+        assert re.match('^[0-9A-Fa-fXx]+$', hexdata), "Invalid hex data. %s" % hexdata
+        data = hex2bytes(hexdata)
+
+        if len(args) == 2:
+            sigs = [args[1]]
+        else:
+            selector = hexdata[:10]
+            sigs = selector_to_sigs(selector, False)
+            if not sigs:
+                print("No selector found for %s." % selector)
+                calldata = hexdata[10:]
+                if not calldata: # No data.
+                    return
+                print("Guessing types ...")
+                i = 0
+                for typ, value in guess_calldata_types(calldata):
+                    print("[%d]   %s   %s" % (i, typ, value))
+                    i += 1
+                return
+            else:
+                print("%s selectors found." % len(sigs))
+        
+        for sig in sigs:
+            if(len(sigs) > 1):
+                print('-----')
+            self.peth.decode_call(sig, data)
+            
 
     def do_common_addresses(self, arg):
         """
@@ -244,6 +319,7 @@ class PethConsole(cmd.Cmd):
             ts = int(arg)
         except Exception as e:
             print("[!] Invalid timestamp number. %s" % e)
+            return
 
         if ts > 3600 * 24 * 365 * 10: 
             # If ts > 10 yrs, it's a timestamp.
@@ -337,6 +413,34 @@ class PethConsole(cmd.Cmd):
     ##################################################################
     # Common used eth_call alias command.
 
+    def do_run(self, arg: str):
+        """
+        run <your_path>.sol : Run the code with eth_call.
+
+        pragma solidity ^0.8.13;
+
+        contract Executor {
+            constructor() public payable {}
+
+            function run() external returns(address, address, address){
+                //////////////////////
+                // ADD YOUR CODE HERE.
+                // It's OK to change the return type or add more functions
+                // or contract. 
+                // But the `Executor` contract and `run()` function must be 
+                // reserved. 
+                //////////////////////
+                return (msg.sender, tx.origin, address(this));
+            }
+        }
+        """
+        assert arg.endswith('.sol'), ".sol file needed but get %s." % arg
+        assert os.path.exists(arg), "File %s not exists." % arg
+        code = open(arg).read()
+        r = self.peth.run_solidity(code)
+        print(r)
+        
+
     def do_get_prop(self, arg):
         """
         get_prop <contract> <name> [<type>] : Call property-like view method.
@@ -536,6 +640,23 @@ class PethConsole(cmd.Cmd):
         print("1 %s = %0.4f %s" % (token0_name, r1/r0, token1_name))
         print("1 %s = %0.4f %s" % (token1_name, r0/r1, token0_name))
 
+
+    def do_factory(self, arg):
+        """
+        factory <factory address>
+        """
+        factory = self.web3.toChecksumAddress(arg)
+        size = self.peth.call_contract(factory, "allPairsLength()->(uint256)")
+        print("%d pairs found." % size)
+        for i in range(size):
+            try:
+                pair = self.peth.call_contract(factory, "allPairs(uint256)->(address)", [i])
+                print("[%d] %s" % (i, pair))
+                self.do_pair(pair)
+            except Exception as e:
+                print('[*] %s' % e)
+
+
     def do_oracle(self, arg):
         """
         oracle <EACAggregatorProxy>[,<EACAggregatorProxy>] : Print ChainLink oracle aggregator information.
@@ -549,14 +670,22 @@ class PethConsole(cmd.Cmd):
                 if cnt > 1:
                     print('---- [%s] ----' % i)
                     i += 1
-
+                
                 aggr = self.peth.call_contract(addr, "aggregator()->(address)")
-                print("Aggregator:", aggr)
+                if aggr is None:
+                    aggr = addr
+                else:
+                    print("Aggregator:", aggr)
+                    print("Proxy Owner:", self.peth.call_contract(
+                    addr, "owner()->(address)"))
+                    
+                print("Aggregator Owner:", self.peth.call_contract(
+                    aggr, "owner()->(address)"))
 
+                print("Version:", self.peth.call_contract(
+                    aggr, "typeAndVersion()->(string)"))
                 print("Description:", self.peth.call_contract(
                     aggr, "description()->(string)"))
-                print("Owner:", self.peth.call_contract(
-                    aggr, "owner()->(address)"))
 
                 dec = self.peth.call_contract(aggr, "decimals()->(uint8)")
                 print("Decimals:", dec)
@@ -610,6 +739,63 @@ class PethConsole(cmd.Cmd):
             data = args[1]
             self.peth.decode_call(sig_or_addr, data)
 
+    def do_idm(self, arg):
+        """
+        idm <addr>  : Print the first 10 IDM messages related to the account.
+        idm <addr> <count>  : Print the first n IDM messages
+        idm <addr> <count> <asc/desc>  : Print the last n IDM messages
+        idm <addr> <count> <asc/desc> <startblock> <endblock> : Print IDM messages between specified blocks.
+        """       
+        args = arg.split()
+        assert len(args) >= 1
+        addr = args[0]
+        count = 10
+        startblock = None
+        endblock = None
+        reverse = False
+
+        if len(args) >= 2:
+            count = int(args[1])
+        if len(args) >= 3:
+            reverse = args[2] == 'desc'
+        if len(args) >= 5:
+            startblock = int(args[3])
+            endblock = int(args[4])
+
+        txs = self.peth.scan.get_txs_by_account(
+            addr, startblock, endblock, count, reverse)
+
+        if not txs:
+            print("No txs.")
+            return
+
+        i = 0
+        for tx in txs:
+
+            sender = tx["from"]
+            to = tx["to"]
+            data = tx["input"]
+            value = int(tx["value"])
+
+            data = data[2:] # remove 0x
+            if len(data) <= 8 or '00' in data:
+                continue
+
+            if sender == addr:
+                sender = "Hacker"
+            if to == addr:
+                to = "Hacker"
+
+            i += 1
+            print("---- [%d] %s to %s (%0.4f)----" % (i, sender, to, value/1e18))
+
+            try:
+                msg = codecs.decode(data, 'hex').decode('utf-8')
+                print(msg)
+            except:
+                print("Decode failed.")
+            
+
     def do_txs(self, arg):
         """
         txs <addr>  : Print the first 10 txs of the account.
@@ -725,7 +911,7 @@ class PethConsole(cmd.Cmd):
         selectors = self.peth.get_selectors(addr)
         for selector in selectors:
             sig = '0x' + selector.hex()
-            sigs = get_4byte_sig(sig)
+            sigs = selector_to_sigs(sig)
             sigs = sigs[::-1]
             print(sig, ', '.join(sigs))
 
@@ -929,16 +1115,18 @@ class PethConsole(cmd.Cmd):
         else:
             print("Nothing downloaded. Check `url {addr}`")
 
-    def do_url(self, addr):
+    def do_url(self, arg):
         """
         url <addr> : Open blockchain explorer of the address.
+        url <tx> : Open blockchain explorer of the tx.
         """
-        if Web3.isAddress(addr):
-            url = self.peth.get_address_url(addr)
-            print(url)
-            self.do_open(url)
+        if Web3.isAddress(arg):
+            url = self.peth.get_address_url(arg)       
         else:
-            print("%s is not a valid address." % addr)
+            if self.peth.address_url:
+                url = self.peth.address_url.replace("address/", "search?f=0&q=") + arg
+        print(url)
+        self.do_open('"%s"' % url)
 
     def do_decompile(self, addr):
         """
